package ac.ed.lurg;

import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import ac.ed.lurg.country.AbstractCountryAgent;
import ac.ed.lurg.country.GlobalPrice;
import ac.ed.lurg.country.StockReader;
import ac.ed.lurg.landuse.CropUsageData;
import ac.ed.lurg.shock.PriceShockManager;
import ac.ed.lurg.types.CropToDoubleMap;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.utils.LogWriter;

public class InternationalMarket {

	private Map<CropType, GlobalPrice> worldPrices;
	private GlobalPrice carbonPrice;
	private GlobalPrice timberPrice;
	private PriceShockManager priceShockManager;
	
	@SuppressWarnings("unchecked")
	public InternationalMarket() {
		if(ModelConfig.IS_CALIBRATION_RUN) {
			worldPrices = new HashMap<CropType, GlobalPrice>();

			Map<CropType, Double> stockLevel = getInitialStockLevels();

			for (CropType crop : CropType.getImportedTypes()) {
				Double initialStock = stockLevel.get(crop);
				worldPrices.put(crop, GlobalPrice.createInitial(crop.getInitialPrice(), initialStock));
			}
			
			carbonPrice = GlobalPrice.createInitial(ModelConfig.CARBON_PRICE, 0);
			timberPrice = GlobalPrice.createInitial(ModelConfig.WOOD_PRICE, 0);
		}
		else {
			List<Object> deserializedPrices = deserializeGlobalPrice();
			worldPrices = (Map<CropType, GlobalPrice>) deserializedPrices.get(0);
			carbonPrice = (GlobalPrice) deserializedPrices.get(1);
			timberPrice = (GlobalPrice) deserializedPrices.get(2);
		}
		priceShockManager = new PriceShockManager();
	}

	public  Map<CropType, GlobalPrice> getWorldPrices() {
		return worldPrices;
	}
	
	public GlobalPrice getCarbonPrice() {
		return carbonPrice;
	}
	
	public GlobalPrice getTimberPrice() {
		return timberPrice;
	}

	private Map<CropType, Double> getInitialStockLevels() {
		Map<CropType, Double> initialStockLevels = new StockReader().read();
		initialStockLevels.put(CropType.ENERGY_CROPS, 0.0);
		return initialStockLevels;
	}

	void determineInternationalTrade(Collection<AbstractCountryAgent> countryAgents, double gen2EcDDemand, double carbonDemand, double timberDemand, Timestep timestep) {
		CropToDoubleMap totalImportCommodities = new CropToDoubleMap();
		CropToDoubleMap totalExportCommodities = new CropToDoubleMap();
		CropToDoubleMap totalProduction = new CropToDoubleMap();
		for (AbstractCountryAgent ca : countryAgents) {

			// Get values for world input costs
			Map<CropType, CropUsageData> cropUsage = ca.getCropUsageData();

			for (Entry<CropType, CropUsageData> entry : cropUsage.entrySet()) {
				CropType c = entry.getKey();

				double countryNetImports = entry.getValue().getShockedNetImports();
				totalProduction.incrementValue(c, (entry.getValue().getProductionExpected()-entry.getValue().getProductionShock()));

				//TODO this doesn't work when both imports and exports > 0
				if (countryNetImports > 0)
					totalImportCommodities.incrementValue(c, countryNetImports);
				else
					totalExportCommodities.incrementValue(c, -countryNetImports);
			}
		}

		// energycrops are a special case where demand in global and exogenously specified
		totalImportCommodities.incrementValue(CropType.ENERGY_CROPS, gen2EcDDemand);

		// Look at trade balance and adjust appropriately
		for (CropType crop : CropType.getImportedTypes()) {

			GlobalPrice prevPrice = worldPrices.get(crop);
			double imports = totalImportCommodities.containsKey(crop) ? totalImportCommodities.get(crop) : 0.0;
			double exportsBeforeTransportLosses = totalExportCommodities.containsKey(crop) ? totalExportCommodities.get(crop) : 0.0;
			LogWriter.println(timestep.getYear() + " Updating " + crop.getGamsName() + " prices");
			GlobalPrice adjustedPrice = prevPrice.createWithUpdatedMarketPrices(imports, exportsBeforeTransportLosses, timestep, totalProduction.get(crop), true);
			LogWriter.println( String.format("Price for %s updated from %s to %s \n", crop.getGamsName(), prevPrice, adjustedPrice));
			if (adjustedPrice.getStockLevel() < 0)
				LogWriter.println("Global stocks are below zero" + crop.getGamsName() + ", " + timestep.getYear());

			worldPrices.put(crop, adjustedPrice);
		}
		
		// Update carbon price
		double totalCarbonSequestered = 0;
		double totalCarbonEmitted = 0;
		for (AbstractCountryAgent ca : countryAgents) {
			double netCarbonFlux = ca.getNetCarbonFlux();
			if (netCarbonFlux >= 0)
				totalCarbonEmitted += netCarbonFlux;
			else
				totalCarbonSequestered -= netCarbonFlux;
		}
		
		totalCarbonSequestered = Math.max(totalCarbonSequestered, 0.0000001); // avoid division by 0
		GlobalPrice prevCPrice = carbonPrice;
		LogWriter.println(timestep.getYear() + " Updating carbon price");
		GlobalPrice adjustedCPrice = prevCPrice.createWithUpdatedMarketPrices(carbonDemand, totalCarbonSequestered, timestep, totalCarbonSequestered, false);
		LogWriter.println( String.format("Price for carbon updated from %s to %s \n", prevCPrice, adjustedCPrice));
		if (adjustedCPrice.getStockLevel() < 0)
			LogWriter.println("Global stocks are below zero carbon, " + timestep.getYear());
		carbonPrice = adjustedCPrice;
		
		// Update timber price
		double totalTimberProduced = 0;
		for (AbstractCountryAgent ca : countryAgents) {
			totalTimberProduced += ca.getTimberProduction();
		}
		totalTimberProduced = Math.max(totalTimberProduced, 0.0000001); // avoid division by 0
		GlobalPrice prevTPrice = timberPrice;
		LogWriter.println(timestep.getYear() + " Updating timber price");
		GlobalPrice adjustedTPrice = prevTPrice.createWithUpdatedMarketPrices(timberDemand, totalTimberProduced, timestep, totalTimberProduced, false);
		LogWriter.println( String.format("Price for timber updated from %s to %s \n", prevTPrice, adjustedTPrice));
		if (adjustedTPrice.getStockLevel() < 0)
			LogWriter.println("Global stocks are below zero timber, " + timestep.getYear());
		timberPrice = adjustedTPrice;
	}

	void writeGlobalMarketFile(Timestep timestep, BufferedWriter outputFile) throws IOException {
		for (CropType crop : CropType.getImportedTypes()) {
			StringBuffer sbData = new StringBuffer();
			GlobalPrice priceQuantity = worldPrices.get(crop);
			sbData.append(String.format("%d,%s", timestep.getYear(), crop.getGamsName()));
			sbData.append(String.format(",%.1f,%.1f", priceQuantity.getImportAmount(), priceQuantity.getExportsAfterTransportLosses()));
			sbData.append(String.format(",%.8f,%.3f", priceQuantity.getExportPrice(), priceQuantity.getStockLevel()));

			outputFile.write(sbData.toString());
			outputFile.newLine();
		}
		// Carbon price
		StringBuffer sbData = new StringBuffer();
		sbData.append(String.format("%d,%s", timestep.getYear(), "carbon"));
		sbData.append(String.format(",%.1f,%.1f", carbonPrice.getImportAmount(), carbonPrice.getExportsAfterTransportLosses()));
		sbData.append(String.format(",%.8f,%.3f", carbonPrice.getExportPrice(), carbonPrice.getStockLevel()));

		outputFile.write(sbData.toString());
		outputFile.newLine();
	}

	public void serializeGlobalPrices() {
		
		List<Object> pricesToSerialize = new ArrayList<Object>();
		pricesToSerialize.add(worldPrices);
		pricesToSerialize.add(carbonPrice);
		pricesToSerialize.add(timberPrice);
		
		try {
			LogWriter.println("Starting serializing GlobalPrice to " + ModelConfig.SERIALIZED_INTERNATIONAL_MARKET_FILE);
			FileOutputStream fileOut = new FileOutputStream(ModelConfig.SERIALIZED_INTERNATIONAL_MARKET_FILE);
			ObjectOutputStream out = new ObjectOutputStream(fileOut);
			out.writeObject(pricesToSerialize);
			out.close();
			fileOut.close();
			LogWriter.println("Serialized international market data is saved");
		} catch (IOException i) {
			i.printStackTrace();
		}
	}
	
	@SuppressWarnings("unchecked")
	private List<Object> deserializeGlobalPrice() {
		try {
			List<Object> initGlobalPrices;
			FileInputStream fileIn = new FileInputStream(ModelConfig.SERIALIZED_INTERNATIONAL_MARKET_FILE);
			ObjectInputStream in = new ObjectInputStream(fileIn);
			initGlobalPrices = (List<Object>) in.readObject();
			in.close();
			fileIn.close();
			LogWriter.println("Deserialized " + ModelConfig.SERIALIZED_INTERNATIONAL_MARKET_FILE);
			return initGlobalPrices;
		} catch (IOException i) {
			LogWriter.printlnError("Problem deserializing " + ModelConfig.SERIALIZED_INTERNATIONAL_MARKET_FILE);
			LogWriter.print(i);
			return null;
		} catch (ClassNotFoundException c) {
			LogWriter.printlnError("GlobalPrice class not found");
			c.printStackTrace();
			return null;
		}
	}
	
	/*
	@SuppressWarnings("unchecked")
	private GlobalPrice deserializeCarbonPrice() {
		try {
			GlobalPrice initCarbonPrice;
			FileInputStream fileIn = new FileInputStream(ModelConfig.SERIALIZED_CARBON_MARKET_FILE);
			ObjectInputStream in = new ObjectInputStream(fileIn);
			initCarbonPrice = (GlobalPrice) in.readObject();
			in.close();
			fileIn.close();
			LogWriter.println("Deserialized " + ModelConfig.SERIALIZED_CARBON_MARKET_FILE);
			return initCarbonPrice;
		} catch (IOException i) {
			LogWriter.printlnError("Problem deserializing " + ModelConfig.SERIALIZED_CARBON_MARKET_FILE);
			LogWriter.print(i);
			return null;
		} catch (ClassNotFoundException c) {
			LogWriter.printlnError("GlobalPrice class not found");
			c.printStackTrace();
			return null;
		}
	}
	*/

	public void applyPriceShocks(Timestep timestep) {
		Map<CropType, Double> shocks = priceShockManager.getShocks(timestep);
		for (Map.Entry<CropType, Double> entry : shocks.entrySet()) {
			CropType crop = entry.getKey();
			GlobalPrice preShock = worldPrices.get(crop);
			double factor = 1.0 + entry.getValue();
			GlobalPrice adjustedPrice = preShock.createPriceAdjustedByFactor(factor);
			LogWriter.println(String.format("applyPriceShocks: %s adjusting price by %.4f to export price %.2f", crop.getFaoName(), factor, adjustedPrice.getExportPrice()));
			worldPrices.put(crop, adjustedPrice);
		}
	}

	/*	public double findMaxPriceDiff(Map<CropType, Double> previousExportPrices) {
		if (previousExportPrices == null)
			return Double.MAX_VALUE;

		double maxSoFar = 0;
		CropType crop = null;

		for (Entry<CropType, GlobalPrice> entry : worldPrices.entrySet()) {
			Double previousP = previousExportPrices.get(entry.getKey());

			if (previousP == null)
				return Double.MAX_VALUE;

			double diffThisTime = Math.abs(previousP - entry.getValue().getExportPrice());
			if (diffThisTime > maxSoFar) {
				maxSoFar = diffThisTime;
				crop = entry.getKey();
			}
		}

		LogWriter.println("findMaxPriceDiff: found " + maxSoFar + " for " + crop);
		return maxSoFar;
	} */
}
