package ac.ed.lurg.country;

import java.io.File;
import java.io.IOException;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.country.gams.GamsCountryInput;
import ac.ed.lurg.country.gams.GamsRasterInput;
import ac.ed.lurg.country.gams.GamsRasterOptimiser;
import ac.ed.lurg.country.gams.GamsRasterOutput;
import ac.ed.lurg.demand.AbstractDemandManager;
import ac.ed.lurg.landuse.CarbonFluxItem;
import ac.ed.lurg.landuse.CropUsageData;
import ac.ed.lurg.landuse.IrrigationItem;
import ac.ed.lurg.landuse.LandUseItem;
import ac.ed.lurg.landuse.WoodYieldItem;
import ac.ed.lurg.landuse.WoodYieldRasterSet;
import ac.ed.lurg.types.CommodityType;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.types.LandCoverType;
import ac.ed.lurg.utils.DoubleMap;
import ac.ed.lurg.utils.LazyHashMap;
import ac.ed.lurg.utils.LazyTreeMap;
import ac.ed.lurg.utils.LogWriter;
import ac.ed.lurg.utils.TripleMap;
import ac.ed.lurg.utils.cluster.Cluster;
import ac.ed.lurg.utils.cluster.KMeans;
import ac.ed.lurg.yield.YieldClusterPoint;
import ac.ed.lurg.yield.YieldRaster;
import ac.ed.lurg.yield.YieldResponsesItem;
import ac.sac.raster.IntegerRasterItem;
import ac.sac.raster.RasterKey;
import ac.sac.raster.RasterSet;

public class CountryAgent extends AbstractCountryAgent {

	private GamsRasterOutput previousGamsRasterOutput;
	private RasterSet<IntegerRasterItem> yieldClusters;
	private Map<CropType, Double> subsidyRates;
	private boolean saveGamsGdxFiles;
	private ForestManager forestManager;

	public CountryAgent(AbstractDemandManager demandManager,CompositeCountry country, RasterSet<LandUseItem> cropAreaRaster,
			Map<CropType, CropUsageData> cropUsageData, Map<CropType, Double> tradeBarriers, RasterSet<IntegerRasterItem> yieldClusters,
			Map<CropType, Double> subsidyRates, WoodYieldRasterSet initWoodYieldData) {

		super(demandManager, country, tradeBarriers);
		this.yieldClusters = yieldClusters;
		this.subsidyRates = subsidyRates;
		
		GamsRasterOutput initialData = new GamsRasterOutput(cropAreaRaster, cropUsageData);
		previousGamsRasterOutput = initialData;
		
		saveGamsGdxFiles = (ModelConfig.GAMS_COUNTRY_TO_SAVE != null && country.getName().equals(ModelConfig.GAMS_COUNTRY_TO_SAVE));
		
		if (yieldClusters != null)
			initialiseForestManager(initWoodYieldData);
		
	}

	public RasterSet<IntegerRasterItem> getYieldClusters() {
		return yieldClusters;
	}

	private RasterSet<IntegerRasterItem> calcYieldClusters(RasterSet<IrrigationItem> irrigData, YieldRaster countryYieldSurfaces) {

		LogWriter.println("calcYieldClusters: for " + ModelConfig.NUM_YIELD_CLUSTERS + " clusters");	
		
		// create collection of ClusteringPoints from countryYieldSurfaces, these have the RasterKey and data for yield (or access to them)
		HashSet<YieldClusterPoint> clusteringPoints = new HashSet<YieldClusterPoint>();
		for (Entry<RasterKey, YieldResponsesItem> entry : countryYieldSurfaces.entrySet()) {
			YieldResponsesItem yieldresp = entry.getValue();
			if (yieldresp != null) {
				RasterKey key = entry.getKey();
				IrrigationItem irrigItem = irrigData.get(key);
				
				LandUseItem luItem = previousGamsRasterOutput.getLandUses().get(key);
				double totalLand = luItem.getTotalLandCoverArea()-luItem.getLandCoverArea(LandCoverType.URBAN);
				double protectedAreaFrac = (totalLand <= 0) ? 0.0 : luItem.getProtectedArea() / totalLand;
				
				clusteringPoints.add(new YieldClusterPoint(key, yieldresp, irrigItem, protectedAreaFrac));
			}
		}
		

		
		// do the clustering
		KMeans<String, YieldClusterPoint> kmeans = new KMeans<String, YieldClusterPoint>(clusteringPoints, ModelConfig.NUM_YIELD_CLUSTERS);
		kmeans.calculateClusters(100, 0.1);
		kmeans.printClusters();

		// reformat output
		List<Cluster<String, YieldClusterPoint>> yieldClusters = kmeans.getClusters();
		RasterSet<IntegerRasterItem> mapping = new RasterSet<IntegerRasterItem>(countryYieldSurfaces.getHeaderDetails());
		
		int id = 1;
		for (Cluster<String, YieldClusterPoint> c : yieldClusters) {
			for (YieldClusterPoint p : c.getPoints())
				mapping.put(p.getRasterKey(), new IntegerRasterItem(id));

			if (c.getPoints().size()>0)
				id++;
		}

		return mapping;
	}
	
	public GamsRasterOutput determineProduction(YieldRaster countryYieldSurfaces, RasterSet<IrrigationItem> irrigData, 
			Map<CropType, GlobalPrice> worldPrices, double globalGen2EcIncrease, RasterSet<CarbonFluxItem> carbonFluxData, 
			RasterSet<WoodYieldItem> woodYieldData, DoubleMap<LandCoverType, LandCoverType, Double> conversionCosts,
			double carbonDemandIncrease, GlobalPrice carbonPrice, double timberDemandIncrease, GlobalPrice timberPrice, WoodYieldRasterSet initWoodYieldData) {

		// project demand
		calculateCountryPricesAndDemand(worldPrices, carbonPrice, timberPrice, false);
		
		if (saveGamsGdxFiles && ModelConfig.PRICE_ELASTIC_DEMAND)
			saveGDXFile("demand");
		
		if (currentProjectedDemand.size() == 0) {
			LogWriter.printlnError("No demand for country " + country + " so skipping it");
		}
		else if (countryYieldSurfaces.size() == 0 ) {
			LogWriter.printlnError("No yield values for " + country + " so skipping it");
		}
		else if (ModelConfig.DEBUG_JUST_DEMAND_OUTPUT) { // if this debug flag is set we don't do the optimisation
		}
		else {			
			if (yieldClusters==null) {
				yieldClusters = calcYieldClusters(irrigData, countryYieldSurfaces);  // this should only be on the first timestep
				initialiseForestManager(initWoodYieldData);
			}
			
			// optimize areas and intensity 
			GamsRasterInput input = getGamsRasterInput(irrigData, countryYieldSurfaces, globalGen2EcIncrease, carbonFluxData, woodYieldData, conversionCosts, 
					carbonDemandIncrease, timberDemandIncrease);
			GamsRasterOptimiser opti = new GamsRasterOptimiser(input, yieldClusters);
			LogWriter.println("Running " + country.getName() + ", currentTimestep " + currentTimestep);

			GamsRasterOutput result = opti.run();

			if (saveGamsGdxFiles)
				saveGDXFile("landuse");
			
			previousGamsRasterOutput = result;
			forestManager.updateForestHistory(result.getNewForest(), result.getReservedDeforested(), currentTimestep);;
			return result;
		}

		throw new RuntimeException("Skipping optimisation of country " + country);
	}
	
	public void updateNetImportsFromProdAndDemand() {
		updateNetImportsFromProdAndDemand(currentProjectedDemand, currentMinDemandFract, previousGamsRasterOutput.getCropUsageData());
	}
	
	private void saveGDXFile(String ext) {
		// some hacky code for debug purposes that keeps each gams gdx file
		try {
			Files.copy(
					FileSystems.getDefault().getPath(ModelConfig.TEMP_DIR + File.separator + "_gams_java_gdb2.gdx"),
					FileSystems.getDefault().getPath(ModelConfig.TEMP_DIR + File.separator + country.getName().replaceAll("\\s+","") + currentTimestep.getYear() + ext + ".gdx"),
					StandardCopyOption.REPLACE_EXISTING);
			Files.copy(
					FileSystems.getDefault().getPath(ModelConfig.TEMP_DIR + File.separator + "_gams_java_gjo1.lst"),
					FileSystems.getDefault().getPath(ModelConfig.TEMP_DIR + File.separator + country.getName().replaceAll("\\s+","") + currentTimestep.getYear() + ext + "Listing" + ".lst"),
					StandardCopyOption.REPLACE_EXISTING);
		} catch (IOException e) {
			LogWriter.print(e);
		}
	}

	private GamsRasterInput getGamsRasterInput(RasterSet<IrrigationItem> irrigData, YieldRaster countryYieldSurfaces, double gen2EcIncrease,
			RasterSet<CarbonFluxItem> carbonFluxData, RasterSet<WoodYieldItem> woodYieldData, DoubleMap<LandCoverType, LandCoverType, Double> conversionCosts,
			double carbonDemandIncrease, double timberDemandIncrease) {
		double allowedImportChange;

		if (currentTimestep.isInitialTimestep() || (ModelConfig.IS_CALIBRATION_RUN && currentTimestep.getTimestep() <= ModelConfig.END_FIRST_STAGE_CALIBRATION)) {  // initialisation time-step
			allowedImportChange = 0.0;
		}
		else { // normal (not the initial) time-step
			allowedImportChange = ModelConfig.MAX_IMPORT_CHANGE;  // when running is calibration model calibrate (ModelConfig.IS_CALIBRATION_RUN==true) MAX_IMPORT_CHANGE is already 0.
		}

		Map<CropType, TradeConstraint> importConstraints = new HashMap<CropType, TradeConstraint>();

		for (Map.Entry<CropType, CropUsageData> entry : previousGamsRasterOutput.getCropUsageData().entrySet()) {
			CropUsageData cropUsage = entry.getValue();
			CropType crop = entry.getKey();
			double baseTrade = cropUsage.getNetImportsExpected();
			double changeUp = 0.0;
			double changeDown = 0.0;

			if (allowedImportChange > 0.0) {
				double maxOfProdOrSupply = cropUsage.getProductionExpected() + Math.max(baseTrade, 0);  //max of supply for food

				// max of supply overall, needed for when imports are supplying feed and change is zero to allow for production to replace imports
		// CONFUSING NOT SURE IF NEEDED		if (maxOfProdOrSupply == 0) maxOfProdOrSupply = Math.max(cropUsage.getNetImports() + cropUsage.getProduction() , cropUsage.getProduction());  
				
				changeDown = changeUp = allowedImportChange * maxOfProdOrSupply;
			}
			// avoid getting stuck at 0
			if (CropType.ENERGY_CROPS.equals(crop)) {
				double croplandArea = LandUseItem.getTotalLandCover(previousGamsRasterOutput.getLandUses().values(), LandCoverType.CROPLAND);
				double ecMaxExport = gen2EcIncrease * croplandArea / 1500.0 * 2; // gen2EcIncrease * croplandArea / totalArea * sharePerCountry
				if (ModelConfig.IS_CALIBRATION_RUN && currentTimestep.getTimestep() <= ModelConfig.END_FIRST_STAGE_CALIBRATION) {
					baseTrade = 0.0;
					ecMaxExport = 0.0; 
				}
				changeDown = ecMaxExport;

				if (Double.isNaN(baseTrade))
					baseTrade=0;
			}

			importConstraints.put(crop, new TradeConstraint(baseTrade - changeDown, baseTrade + changeUp));
		}
		
		// Carbon import/export constraints TODO not used
		TradeConstraint carbonTradeConstraint;
		{
			double baseTrade = getNetCarbonFlux();
			double countryArea = LandUseItem.getTotalLandArea(previousGamsRasterOutput.getLandUses().values());
			double change = 0.0;
			if (allowedImportChange > 0.0)
				change = Math.max(baseTrade * allowedImportChange, carbonDemandIncrease * countryArea / 1500 * 2);
			carbonTradeConstraint = new TradeConstraint(baseTrade - change, baseTrade + change);
		}

		// Timber import/export constraints
		TradeConstraint timberTradeConstraint;
		{
			double baseTrade = getTimberProduction();
			double countryArea = LandUseItem.getTotalLandArea(previousGamsRasterOutput.getLandUses().values());
			double change = 0.0;
			if (allowedImportChange > 0.0)
				change = Math.max(baseTrade * allowedImportChange, timberDemandIncrease * countryArea / 1500 * 2);

			timberTradeConstraint = new TradeConstraint(baseTrade - change, baseTrade + change);
		}
		
		// Aggregate minimum land cover for current timestep
		DoubleMap<Integer, LandCoverType, Double> minimumLandCoverForTimestep = forestManager.getReservedAreaForTimestep(currentTimestep); //<Location, LandCoverType, Area>
		DoubleMap<Integer, LandCoverType, Double> timberYieldForTimestep = forestManager.getTimberYieldForTimestep(currentTimestep); //<Location, LandCoverType, Area>
		Map<Integer, Double> naturalYieldFactorForTimestep = forestManager.getNaturalYieldFactorForTimestep(currentTimestep);

		GamsCountryInput countryLevelInputs = new GamsCountryInput(country, currentProjectedDemand, currentCountryPrices, importConstraints, 
				previousGamsRasterOutput.getCropUsageData(), currentMinDemandFract, subsidyRates, currentCarbonPrice, carbonTradeConstraint, currentTimberPrice, timberTradeConstraint);	
		GamsRasterInput input = new GamsRasterInput(currentTimestep, countryYieldSurfaces, previousGamsRasterOutput.getLandUses(), irrigData, 
				carbonFluxData, woodYieldData, minimumLandCoverForTimestep, timberYieldForTimestep, naturalYieldFactorForTimestep, conversionCosts, countryLevelInputs);

		return input;
	}
	
	@Override
	protected double getCommPriceFromCropPrice(CommodityType commodity) {
		double commPricePlum = 0;
		Map<CropType, Double> demandFract = baseDemandFact.get(commodity);
		Map<CropType, CropUsageData> cropUsageMap = previousGamsRasterOutput.getCropUsageData();
		
		for (CropType crop : commodity.getCropTypes()) {
			CropUsageData cropUsageData = cropUsageMap.get(crop);
	
			double weighting = Math.max(0, cropUsageData.getShockedNetImports()/(cropUsageData.getProductionExpected()+cropUsageData.getNetImportsExpected()));
			
			double importPrice = currentCountryPrices.get(crop).getImportPrice();
			double exportPrice = currentCountryPrices.get(crop).getExportPrice();
			double prodCost = cropUsageData.getProdCostRate();
			double exportOrProdCost = Double.isNaN(prodCost) || prodCost <= 0.0 ? exportPrice : prodCost;
			double newCropPrice;
			
            switch (ModelConfig.PRICE_CALCULATION) {
                case IMPORTONLY:
                	newCropPrice = importPrice;
                    break;
                case WEIGHTEDIMPEXP:
                	newCropPrice = importPrice*weighting + exportPrice*(1-weighting);
                    break;
                case WEIGHTEDINCLPRODCOSTS:
                	newCropPrice = importPrice*weighting + exportOrProdCost*(1-weighting);
                	break;
                default:
                	newCropPrice = importPrice;
                	LogWriter.printlnWarning("Warning, price calculation type not set, using only import price as default " + newCropPrice);
                    break;
            }
	
			LogWriter.println(String.format("   Crop %s: import %.4f, export %.4f, prodcost %.4f, newCropPrice %.4f", crop, importPrice, exportPrice, prodCost, newCropPrice));
			
			commPricePlum += newCropPrice * demandFract.get(crop);  // weight price by base demand of each cereal crop
		}
		
		return commPricePlum;
	}
	
	public void shockGDP() {

		double totalProduction = getTotalProduction();
		double shockMagnitude = 0.0;

		for (Map.Entry<CropType, CropUsageData> entry : previousGamsRasterOutput.getCropUsageData().entrySet()) {
			CropUsageData cropUsage = entry.getValue();
			CropType crop = entry.getKey();
			if(crop.isImportedCrop() && cropUsage.getProductionExpected() != 0) 
				shockMagnitude += (cropUsage.getProductionShock() /cropUsage.getProductionExpected()) * (cropUsage.getProductionExpected()/totalProduction);	
		}
		demandManager.updateGdpLossesFromShock(country, currentTimestep.getYear(), shockMagnitude);
	}

	public RasterSet<LandUseItem> getLandUses() {
		return previousGamsRasterOutput.getLandUses();
	}

	public Map<CropType, CropUsageData> getCropUsageData() {
		return previousGamsRasterOutput.getCropUsageData();
	}
	
	public double getTotalProduction() {
		double totalProduction = 0;
		
		for (Map.Entry<CropType, CropUsageData> entry : previousGamsRasterOutput.getCropUsageData().entrySet()) {
			CropUsageData cropUsage = entry.getValue();
			CropType crop = entry.getKey();
			if(crop.isImportedCrop()) //assuming pasture and set aside not part of agricultural value, valid or not?
				totalProduction += cropUsage.getProductionExpected();
		}
		return totalProduction;
	}
	
	public void updateProtectedAreas(RasterSet<LandUseItem> newProtectedAreas) {
		for (Entry<RasterKey, LandUseItem> entry : previousGamsRasterOutput.getLandUses().entrySet()) {
			double updatedPA = newProtectedAreas.get(entry.getKey()).getProtectedArea();
			entry.getValue().setProtectedArea(updatedPA);
		}
		
	}

	private void initialiseForestManager(WoodYieldRasterSet initWoodYieldData) {
		forestManager = new ForestManager();
		// Add initial natural areas and managed forest. Assume that unmanaged forest is a mature natural area (climax community) and other natural is a new natural area
		forestManager.initialiseForestHistory(yieldClusters, previousGamsRasterOutput.getLandUses(), initWoodYieldData);
	}
	
	public double getNetCarbonFlux() {
		return previousGamsRasterOutput.getNetCarbonFlux();
	}
	
	public double getTimberProduction() {
		return previousGamsRasterOutput.getTimberProduction();
	}
}