package ac.ed.lurg.country;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import ac.ed.lurg.InternationalMarket;
import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.Timestep;
import ac.ed.lurg.country.crafty.CraftyCountryAgent;
import ac.ed.lurg.country.crafty.CraftyProdManager;
import ac.ed.lurg.demand.AbstractDemandManager;
import ac.ed.lurg.landuse.CarbonFluxItem;
import ac.ed.lurg.landuse.CarbonFluxRasterSet;
import ac.ed.lurg.landuse.CropUsageData;
import ac.ed.lurg.landuse.IrrigationItem;
import ac.ed.lurg.landuse.IrrigationRasterSet;
import ac.ed.lurg.landuse.LandUseItem;
import ac.ed.lurg.landuse.WoodYieldItem;
import ac.ed.lurg.landuse.WoodYieldRasterSet;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.types.LandCoverType;
import ac.ed.lurg.utils.DoubleMap;
import ac.ed.lurg.utils.LogWriter;
import ac.ed.lurg.yield.YieldRaster;
import ac.sac.raster.IntegerRasterItem;
import ac.sac.raster.RasterKey;
import ac.sac.raster.RasterSet;

public class CountryAgentManager {
	private AbstractDemandManager demandManager;
	private TradeManager tradeBarrierManager;
	private SubsidyRateManager subsidyRateManager;
	private InternationalMarket internationalMarket;
	private CountryBoundaryRaster countryBoundaryRaster;
	private RasterSet<IntegerRasterItem> clusterIdRaster;
	private RasterSet<LandUseItem> globalLandUseRaster;
	private WoodYieldRasterSet initWoodYieldData;
	
	private Collection<AbstractCountryAgent> countryAgents = new ArrayList<AbstractCountryAgent>();
	private Collection<CountryAgent> gamsCountryAgents = new ArrayList<CountryAgent>();
	private Collection<CraftyCountryAgent> craftyCountryAgents = new ArrayList<CraftyCountryAgent>();
	private CraftyProdManager craftyManager;
	
	public CountryAgentManager (CompositeCountryManager compositeCountryManager, AbstractDemandManager demandManager, 
			CountryBoundaryRaster countryBoundaryRaster, InternationalMarket internationalMarket,
			RasterSet<IntegerRasterItem> clusterIdRaster, RasterSet<LandUseItem> globalLandUseRaster, WoodYieldRasterSet initWoodYieldData) {
		this.demandManager = demandManager;
		this.countryBoundaryRaster = countryBoundaryRaster;
		this.clusterIdRaster = clusterIdRaster;
		this.internationalMarket = internationalMarket;
		this.globalLandUseRaster = globalLandUseRaster;
		this.initWoodYieldData = initWoodYieldData;
		tradeBarrierManager = new TradeManager(compositeCountryManager);
		subsidyRateManager = new SubsidyRateManager(compositeCountryManager);
		craftyManager = new CraftyProdManager();
	}
	
	public void addForCountry(CompositeCountry cc, Map<CompositeCountry, Map<CropType, CropUsageData>> cropUsageDataMap, RasterSet<LandUseItem> initLU) {
		if (ModelConfig.DEBUG_LIMIT_COUNTRIES) { // DEBUG code
			if (!(cc.getName().equals(ModelConfig.DEBUG_COUNTRY_NAME)))
				return;
		}

		Map<CropType, Double> tradeBarriers = tradeBarrierManager.getTradeBarriers(cc);
		
		if (ModelConfig.USE_CRAFTY_COUNTRIES && craftyManager.getCraftyCountries().contains(cc)) {
			LogWriter.println("Creating CRAFTY agent for: " + cc);
			CraftyCountryAgent cca = new CraftyCountryAgent(demandManager, cc, tradeBarriers);
			craftyCountryAgents.add(cca);
			countryAgents.add(cca);
		}
		else { //GAMS
			LogWriter.println("Creating GAMS agent for: " + cc);
			List<RasterKey> keys = countryBoundaryRaster.getKeysFor(cc);
			Map<CropType, CropUsageData> countryCommodityData = cropUsageDataMap.get(cc);
	
			if (countryCommodityData == null) {
				LogWriter.printlnError("No commodities data for " + cc + ", so skipping");
			} else {
				RasterSet<LandUseItem> initCountryLandUse = initLU.createSubsetForKeys(keys);
				RasterSet<IntegerRasterItem> yieldClusters = ModelConfig.GENERATE_NEW_YIELD_CLUSTERS ? null : clusterIdRaster.createSubsetForKeys(keys);
	
				if (initCountryLandUse.size() == 0) {
					LogWriter.printlnError("No initial land use for " + cc + ", so skipping");
				}
				else {
					Map<CropType, Double> subsidyRates = subsidyRateManager.getSubsidyRates(cc);
					
					CountryAgent ca = new CountryAgent(demandManager, cc, initCountryLandUse, countryCommodityData, tradeBarriers, yieldClusters, subsidyRates, initWoodYieldData);
					gamsCountryAgents.add(ca);
					countryAgents.add(ca);
				}
			}
		}
	}
	
	public Collection<AbstractCountryAgent> getAll() {
		return countryAgents;
	}

	public void determineProductionForAll(Timestep timestep, YieldRaster yieldSurfaces, IrrigationRasterSet currentIrrigationData, double gen2Increase,
			CarbonFluxRasterSet currentCarbonFluxData, WoodYieldRasterSet currentWoodYieldData, DoubleMap<LandCoverType, LandCoverType, Double> conversionCosts,
			double carbonDemandIncrease, double timberDemandIncrease) {
		
		for (AbstractCountryAgent aca : countryAgents) {		
			aca.setCurrentTimestep(timestep);
		}
		
		for (CountryAgent ca : gamsCountryAgents) {		
			LogWriter.println("Country " + ca.getCountry());
			Collection<RasterKey> countryKeys = countryBoundaryRaster.getKeysFor(ca.getCountry());
			YieldRaster countryYieldSurfaces = yieldSurfaces.createSubsetForKeys(countryKeys);
			RasterSet<IrrigationItem> irrigData = currentIrrigationData.createSubsetForKeys(countryKeys);
			RasterSet<CarbonFluxItem> carbonFluxData = currentCarbonFluxData.createSubsetForKeys(countryKeys);
			RasterSet<WoodYieldItem> woodYieldData = currentWoodYieldData.createSubsetForKeys(countryKeys);
			
			try {
				ca.determineProduction(countryYieldSurfaces, irrigData, internationalMarket.getWorldPrices(), gen2Increase, carbonFluxData, woodYieldData, 
						conversionCosts, carbonDemandIncrease, internationalMarket.getCarbonPrice(), timberDemandIncrease, internationalMarket.getTimberPrice(), initWoodYieldData);
				
				// update global rasters
				globalLandUseRaster.putAll(ca.getLandUses());

				// if first timestep and calibration get the clustering info, which doesn't change through time
				if (ModelConfig.GENERATE_NEW_YIELD_CLUSTERS && timestep.isInitialTimestep())
					clusterIdRaster.putAll(ca.getYieldClusters());
				
			} catch (Exception e) {
 				LogWriter.printlnError("Exception processing " + ca.getCountry() + " will continue with other countries");
				LogWriter.print(e);
			}
		}
		
		if (craftyCountryAgents.size() > 0) {
			craftyManager.updateWithCraftyData(craftyCountryAgents, timestep, internationalMarket.getWorldPrices(), 
					internationalMarket.getCarbonPrice(), internationalMarket.getTimberPrice());  // this will wait for the marker file from CRAFTY
		}
	}
	
	public void updateNetImportsForAll() {
		for (CountryAgent ca : gamsCountryAgents) {
			ca.updateNetImportsFromProdAndDemand();
		}
	}

	public void recalculateDemandForAll() {
		for (CountryAgent ca : gamsCountryAgents) {
			ca.shockGDP();
			ca.calculateCountryPricesAndDemand(internationalMarket.getWorldPrices(), internationalMarket.getCarbonPrice(), internationalMarket.getTimberPrice(), true);
		}
	}
	
	public void updateProtectedAreasForAll(RasterSet<LandUseItem> newProtectedAreas) {
		for (CountryAgent ca : gamsCountryAgents) {
			Collection<RasterKey> countryKeys = countryBoundaryRaster.getKeysFor(ca.getCountry());
			RasterSet<LandUseItem> newPA = newProtectedAreas.createSubsetForKeys(countryKeys);
			ca.updateProtectedAreas(newPA);
		}
	}
	
	public void serializeCropUsageForAll() {
		Map<CompositeCountry, Map<CropType, CropUsageData>> cropUsageDataMap = new HashMap<CompositeCountry, Map<CropType, CropUsageData>>();
		for (CountryAgent ca : gamsCountryAgents) {
			cropUsageDataMap.put(ca.country, ca.getCropUsageData());
		}
		
		try {
			LogWriter.println("Starting serializing CropUsages to " + ModelConfig.SERIALIZED_CROP_USAGE_FILE);
			FileOutputStream fileOut = new FileOutputStream(ModelConfig.SERIALIZED_CROP_USAGE_FILE);
			ObjectOutputStream out = new ObjectOutputStream(fileOut);
			out.writeObject(cropUsageDataMap);
			out.close();
			fileOut.close();
			LogWriter.println("Serialized crop usage data is saved");
		} catch (IOException i) {
			i.printStackTrace();
		}
		
		
	}	
	
}