package ac.ed.lurg.country;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.Timestep;
import ac.ed.lurg.landuse.LandUseItem;
import ac.ed.lurg.landuse.WoodYieldRasterSet;
import ac.ed.lurg.types.LandCoverType;
import ac.ed.lurg.utils.DoubleMap;
import ac.ed.lurg.utils.LazyTreeMap;
import ac.ed.lurg.utils.LogWriter;
import ac.sac.raster.IntegerRasterItem;
import ac.sac.raster.RasterKey;
import ac.sac.raster.RasterSet;
import ac.ed.lurg.landuse.WoodYieldItem;

public class ForestManager {
	private List<ForestRecord> forestHistory;
	
	ForestManager() {
		forestHistory = new ArrayList<ForestRecord>();
	}
	
	public void addForestRecord(ForestRecord record) {
		forestHistory.add(new ForestRecord(record.getYearPlanted(), record.getLocation(), record.getForestType(), record.getArea(), record.getPotentialYield()));	
	}
	
	public void initialiseForestHistory(RasterSet<IntegerRasterItem> yieldClusters, RasterSet<LandUseItem> initialLandUse, WoodYieldRasterSet initWoodYieldData) {
		LazyTreeMap<Integer, Double> unmanagedForestArea = new LazyTreeMap<Integer, Double>() {
			private static final long serialVersionUID = 1L;
			protected Double createValue() {return 0.0;}
		};
		LazyTreeMap<Integer, Double> otherNaturalArea = new LazyTreeMap<Integer, Double>() {
			private static final long serialVersionUID = 1L;
			protected Double createValue() {return 0.0;}
		};
		LazyTreeMap<Integer, Double> timberForestArea = new LazyTreeMap<Integer, Double>() {
			private static final long serialVersionUID = 1L;
			protected Double createValue() {return 0.0;}
		};
		LazyTreeMap<Integer, Double> timberYieldTimesArea = new LazyTreeMap<Integer, Double>() {
			private static final long serialVersionUID = 1L;
			protected Double createValue() {return 0.0;}
		};
		
		for (Entry<RasterKey, LandUseItem> entry : initialLandUse.entrySet()) {
			RasterKey key = entry.getKey();
			LandUseItem luItem = entry.getValue();
			if (luItem == null) {
				LogWriter.printlnWarning("CountryAgent.initialiseForestManager: null luItem");
				continue;
			}
			try {
				int clusterId = yieldClusters.get(key).getInt();

				double unmanagedForestAreaSoFar = unmanagedForestArea.lazyGet(clusterId);
				double otherNaturalAreaSoFar = otherNaturalArea.lazyGet(clusterId);
				double timberForestAreaSoFar = timberForestArea.lazyGet(clusterId);
				double timberYieldSoFar = timberYieldTimesArea.lazyGet(clusterId);
				//LogWriter.println(key.toString());
				WoodYieldItem wYItem = initWoodYieldData.get(key);
				double timberYieldThisTime;
				timberYieldThisTime = (wYItem != null) ? initWoodYieldData.get(key).getWoodYield(LandCoverType.PASTURE, LandCoverType.TIMBER_FOREST) : 0.0;
		
				unmanagedForestArea.put(clusterId, unmanagedForestAreaSoFar + luItem.getInitialForestedNaturalArea());
				otherNaturalArea.put(clusterId, otherNaturalAreaSoFar + luItem.getLandCoverArea(LandCoverType.NATURAL) - luItem.getInitialForestedNaturalArea()); // Natural is other natural + unmanaged forest so need to subtract the latter
				timberForestArea.put(clusterId, timberForestAreaSoFar + luItem.getLandCoverArea(LandCoverType.TIMBER_FOREST));
				timberYieldTimesArea.put(clusterId, timberYieldSoFar + luItem.getLandCoverArea(LandCoverType.TIMBER_FOREST) * timberYieldThisTime);
				
				/*
				if (clusterId == 3) {
					LogWriter.println(""+timberYieldThisTime+","+timberYieldSoFar+","+luItem.getLandCoverArea(LandCoverType.TIMBER_FOREST)+","+timberForestAreaSoFar);
				}*/

			} catch (Exception e) {
				LogWriter.print(e);
			}
		}
		
		// Assume unmanaged forest is fully mature (max wood yield) and other natural is new (no wood yield).
		int otherNaturalPseudoYear = ModelConfig.BASE_YEAR;
		int forestPseudoYear = ModelConfig.BASE_YEAR - ModelConfig.FOREST_LOCKIN_PERIOD;
		
		for (Entry<Integer, Double> entry : unmanagedForestArea.entrySet()) {
			addForestRecord(new ForestRecord(forestPseudoYear, entry.getKey(), LandCoverType.NATURAL, entry.getValue(), 0));
		}
		
		for (Entry<Integer, Double> entry : otherNaturalArea.entrySet()) {
			addForestRecord(new ForestRecord(otherNaturalPseudoYear, entry.getKey(), LandCoverType.NATURAL, entry.getValue(), 0));
		}
		
		for (Entry<Integer, Double> entry : timberForestArea.entrySet()) {
			int clusterId = entry.getKey();
			double area = entry.getValue();
			double timberYield = (area > 0) ? timberYieldTimesArea.get(clusterId) / area : 0.0;
			addForestRecord(new ForestRecord(forestPseudoYear, entry.getKey(), LandCoverType.TIMBER_FOREST, area, timberYield));
		}
	}
	
	public DoubleMap<Integer, LandCoverType, Double> getReservedAreaForTimestep(Timestep timestep) {
		DoubleMap<Integer, LandCoverType, Double> minimumLandCoverForTimestep = new DoubleMap<Integer, LandCoverType, Double>();
		double minYear = timestep.getYear() - ModelConfig.FOREST_LOCKIN_PERIOD;
		for (ForestRecord record : forestHistory) {
			if (record.getYearPlanted() > minYear && record.getForestType().isManagedForest()) {
				minimumLandCoverForTimestep.addTo(record.getLocation(), record.getForestType(), record.getArea());
			}
		}
		
		return minimumLandCoverForTimestep;
	}
	
	public DoubleMap<Integer, LandCoverType, Double> getTimberYieldForTimestep(Timestep timestep) {
		DoubleMap<Integer, LandCoverType, Double> yieldTimesArea = new DoubleMap<Integer, LandCoverType, Double>();
		DoubleMap<Integer, LandCoverType, Double> area = new DoubleMap<Integer, LandCoverType, Double>();
		double yearPlanted = timestep.getYear() - ModelConfig.FOREST_LOCKIN_PERIOD;
		for (ForestRecord record : forestHistory) {
			if (record.getYearPlanted() == yearPlanted && record.getForestType().isManagedForest()) {
				yieldTimesArea.addTo(record.getLocation(), record.getForestType(), record.getPotentialYield() * record.getArea());
				area.addTo(record.getLocation(), record.getForestType(), record.getArea());
			}
		}
		
		DoubleMap<Integer, LandCoverType, Double> yield = new DoubleMap<Integer, LandCoverType, Double>();
		// Weighted average
		for (Entry<Integer, Map<LandCoverType, Double>> locMap : yieldTimesArea.getMap().entrySet()) {
			for (Entry<LandCoverType, Double> coverMap : locMap.getValue().entrySet()) {
				double ya = yieldTimesArea.get(locMap.getKey(), coverMap.getKey());
				double a = area.get(locMap.getKey(), coverMap.getKey());
				double y = (a > 0) ? ya / a : a;
				yield.put(locMap.getKey(), coverMap.getKey(), y);
			}
		}
		
		return yield;
	}
	
	// Calculates wood yield multiplier for natural areas based on age
	double naturalWoodYieldFunction(double age) {
		return Math.min(age, ModelConfig.FOREST_LOCKIN_PERIOD) / ModelConfig.FOREST_LOCKIN_PERIOD;
	}
	
	public Map<Integer, Double> getNaturalYieldFactorForTimestep(Timestep timestep) {
		DoubleMap<Integer, LandCoverType, Double> ageTimesArea = new DoubleMap<Integer, LandCoverType, Double>();
		DoubleMap<Integer, LandCoverType, Double> area = new DoubleMap<Integer, LandCoverType, Double>();
		int currentYear = timestep.getYear();
		for (ForestRecord record : forestHistory) {
			if (record.getForestType().equals(LandCoverType.NATURAL)) {
				ageTimesArea.addTo(record.getLocation(), record.getForestType(), record.getArea() * (currentYear - record.getYearPlanted()));
				area.addTo(record.getLocation(), record.getForestType(), record.getArea());
			}
		}
		
		Map<Integer, Double> yieldFactor = new HashMap<Integer, Double>();
		// Weighted average
		for (Entry<Integer, Map<LandCoverType, Double>> locMap : ageTimesArea.getMap().entrySet()) {
			for (Entry<LandCoverType, Double> coverMap : locMap.getValue().entrySet()) {
				double y = ageTimesArea.get(locMap.getKey(), coverMap.getKey());
				double a = area.get(locMap.getKey(), coverMap.getKey());
				if (a != 0) {
					yieldFactor.put(locMap.getKey(), naturalWoodYieldFunction(y / a));
				} else {
					yieldFactor.put(locMap.getKey(), 0.0);
				}
				
			}
		}

		return yieldFactor;	
	}

	public double getForestPlantedArea(int year, int location, LandCoverType forestType) {
		List<ForestRecord> recordsFound = forestHistory
				.stream()
				.filter(r -> r.getYearPlanted() == year && r.getLocation() == location && r.getForestType() == forestType)
				.collect(Collectors.toList());
		
		if (recordsFound.size() > 1)
			LogWriter.printlnError("ForestManager.getForestPlantedArea(): duplicate records found.");
		
		if (recordsFound.isEmpty())
			return 0.0;
		
		return recordsFound.get(0).getArea();
	}
	
	public void setForestPlantedArea(int year, int location, LandCoverType forestType, double area) {
		List<ForestRecord> recordsFound = forestHistory
				.stream()
				.filter(r -> r.getYearPlanted() == year && r.getLocation() == location && r.getForestType() == forestType)
				.collect(Collectors.toList());
		
		if (recordsFound.size() > 1)
			LogWriter.printlnError("ForestManager.getForestPlantedArea(): duplicate records found.");
		
		if (recordsFound.isEmpty())
			LogWriter.printlnError("ForestManager.getForestPlantedArea(): no records found.");
		
		recordsFound.get(0).setArea(area);
	}
	
	public double getPotentialYield(int year, int location, LandCoverType forestType) {
		List<ForestRecord> recordsFound = forestHistory
				.stream()
				.filter(r -> r.getYearPlanted() == year && r.getLocation() == location && r.getForestType() == forestType)
				.collect(Collectors.toList());
		
		if (recordsFound.size() > 1)
			LogWriter.printlnError("ForestManager.getPotentialYield(): duplicate records found.");
		
		if (recordsFound.isEmpty())
			LogWriter.printlnError("ForestManager.getPotentialYield(): no records found.");
		
		return recordsFound.get(0).getPotentialYield();
	}

	public void updateForestHistory(List<ForestRecord> newForest, DoubleMap<Integer, LandCoverType, Double> reservedDeforested, Timestep timestep) {
		int yearPlanted = timestep.getYear();
		for (ForestRecord item : newForest) {
			forestHistory.add(new ForestRecord(yearPlanted, item));
		}

		// Remove deforested area from minimum area requirement
		if (reservedDeforested.getMap().isEmpty())
			return;

		for (Entry<Integer, Map<LandCoverType, Double>> locMap : reservedDeforested.getMap().entrySet()) {
			int locId = locMap.getKey();

			for (Entry<LandCoverType, Double> coverMap : locMap.getValue().entrySet()) {
				LandCoverType forestType = coverMap.getKey();
				double deforestedArea = coverMap.getValue();
				List<Integer> yearsFound = new ArrayList<Integer>(); 
				for (ForestRecord item : forestHistory) {
					yearsFound.add(item.getYearPlanted());
				}
				Collections.sort(yearsFound, Collections.reverseOrder()); // descending order, remove area from newest forest first
				int minYear = timestep.getYear() - ModelConfig.FOREST_LOCKIN_PERIOD;
				
				for (int year : yearsFound) {
					if (year > minYear) {
						double currentArea = getForestPlantedArea(year, locId, forestType);
						if (currentArea < 0.0) 
							LogWriter.printlnError("ForestManager.updateForestHistory(): currentMinimumArea < 0");
						else if (currentArea == 0.0) {
							continue;
						}

						double areaToSubtract = Math.min(currentArea, deforestedArea);
						if (currentArea - areaToSubtract < 0.0) { // floating point errors
							areaToSubtract = currentArea;
						}

						setForestPlantedArea(year, locId, forestType, currentArea - areaToSubtract);
						deforestedArea -= areaToSubtract;
						if (deforestedArea < 1E-6) 
							break;
					}
				}

				if (deforestedArea >= 1E-6)
					LogWriter.printlnError("ForestManager.updateForestHistory(): unable to subtract all deforested area");
			}
		} 
	}
}
