package ac.ed.lurg.country.gams;

import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.landuse.CarbonFluxItem;
import ac.ed.lurg.landuse.Intensity;
import ac.ed.lurg.landuse.IrrigationItem;
import ac.ed.lurg.landuse.LandUseItem;
import ac.ed.lurg.landuse.WoodYieldItem;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.types.LandCoverType;
import ac.ed.lurg.types.YieldType;
import ac.ed.lurg.utils.DoubleMap;
import ac.ed.lurg.utils.LazyTreeMap;
import ac.ed.lurg.utils.LogWriter;
import ac.ed.lurg.yield.YieldRaster;
import ac.ed.lurg.yield.YieldResponsesItem;
import ac.sac.raster.IntegerRasterItem;
import ac.sac.raster.RasterKey;
import ac.sac.raster.RasterSet;

public class GamsRasterOptimiser {
	public static final boolean DEBUG = false;

	private GamsRasterInput rasterInputData;
	private RasterSet<IntegerRasterItem> mapping;

	public GamsRasterOptimiser(GamsRasterInput rasterInputData, RasterSet<IntegerRasterItem> clusterMapping) {
		this.rasterInputData = rasterInputData;
		this.mapping = clusterMapping;
	}

	public GamsRasterOutput run() {
		// workout similar areas		
		GamsLocationInput gamsInput = convertFromRaster(rasterInputData);

		long start = System.currentTimeMillis();

		GamsLocationOptimiser opti = new GamsLocationOptimiser(gamsInput);		
		GamsLocationOutput gamsOutput = opti.run();

		long last = System.currentTimeMillis();
		LogWriter.println("Took " + (last-start));

		// map results back to raster
		return convertToRaster(gamsInput, gamsOutput);
	}

	private GamsRasterOutput convertToRaster(GamsLocationInput gamsInput, GamsLocationOutput gamsOutput) {		
		RasterSet<LandUseItem> newIntensityRaster = allocAreas(gamsInput.getPreviousLandUse(), gamsOutput, gamsInput.getTimestep().getYear());

		return new GamsRasterOutput(gamsOutput.getStatus(), newIntensityRaster, gamsOutput.getCommoditiesData(), gamsOutput.getNewForest(), 
				gamsOutput.getReservedDeforested(), gamsOutput.getNetCarbonFlux(), gamsOutput.getTimberProduction());
	}

	private RasterSet<LandUseItem> createWithSameLandCovers(RasterSet<LandUseItem> toCopy) {
		RasterSet<LandUseItem> theCopy = new RasterSet<LandUseItem>(toCopy.getHeaderDetails());

		for (Entry<RasterKey, LandUseItem> entry : toCopy.entrySet()) {
			if (entry.getValue() != null) {
				LandUseItem newAreasItem =  new LandUseItem(entry.getValue());
				theCopy.put(entry.getKey(), newAreasItem);
			}
		}

		return theCopy;
	}

	private void checkedTotalAreas(Map<? extends Object, LandUseItem> areaRaster, String comment) {
		for (LandCoverType l : LandCoverType.values()) {
			double total = 0;
			double unprotected = 0;
			for (LandUseItem a : areaRaster.values()) {
				if (a != null) {
					total += a.getLandCoverArea(l);
					unprotected += a.getUnprotectedLandCoverArea(l);
				}
										
			}

			LogWriter.println("Total Area " + comment + ": " + l.getName() + ": total = " + total + ", unprotected = " + unprotected);
		}

		double suitableArea=0, protectedArea=0;
		for (LandUseItem a : areaRaster.values()) {
			if (a != null) {
				protectedArea += a.getProtectedArea();
				suitableArea += a.getSuitableArea();
			}
		}
		LogWriter.println("Total protectedArea " + comment + ": " + protectedArea);
		LogWriter.println("Total suitableArea " + comment + ": " + suitableArea);
	}

	private RasterSet<LandUseItem> allocAreas(Map<Integer, ? extends LandUseItem> prevAreasAgg, GamsLocationOutput gamsOutput, int year) {
		RasterSet<LandUseItem> newLandUseRaster = createWithSameLandCovers(rasterInputData.getPreviousLandUses());

		for (Map.Entry<Integer, LandUseItem> entry : gamsOutput.getLandUses().entrySet()) {
			Integer locId = entry.getKey();
			LandUseItem newLandUseAggItem = entry.getValue();

			Set<RasterKey> keys = new HashSet<RasterKey>();
			for (Entry<RasterKey, IntegerRasterItem> mapEntry : mapping.entrySet()) {
				IntegerRasterItem iri = mapEntry.getValue();

				if (iri != null && locId == iri.getInt())
					keys.add(mapEntry.getKey());
			}
			
			RasterSet<LandUseItem> landUseItemsForLocation = newLandUseRaster.createSubsetForKeys(keys);
			
			DoubleMap<LandCoverType, LandCoverType, Double> landCoverChange = gamsOutput.getLandCoverChanges().getInnerMap(locId);
			
			// Allocate land cover change to grid cells
			for (Entry<LandCoverType, Map<LandCoverType, Double>> fromMap: landCoverChange.getMap().entrySet()) {
				LandCoverType fromLC = fromMap.getKey();
				for (Entry<LandCoverType, Double> toMap: fromMap.getValue().entrySet()) {
					LandCoverType toLC = toMap.getKey();
					double change = toMap.getValue();
					allocAllLandCropsTop(newLandUseRaster, keys, toLC, fromLC, change, locId);
				}
			}

			for (RasterKey key : keys) {
				LandUseItem newLandUseItem = newLandUseRaster.get(key);
				if (newLandUseItem != null) {
					for (CropType crop : CropType.values()) {
						newLandUseItem.setCropFraction(crop, newLandUseAggItem.getCropFraction(crop));
						newLandUseItem.setIntensity(crop, newLandUseAggItem.getIntensity(crop)); // intensities constant over single aggregated land category
					}
				}
			}
			
			if (DEBUG) checkedTotalAreas(landUseItemsForLocation, locId + " after");
		}

		return newLandUseRaster;
	}

	private void allocAllLandCropsTop(RasterSet<LandUseItem> newLandUseRaster, Set<RasterKey> keys, LandCoverType toLC, LandCoverType fromLC, double change, Integer locId) {
		LandCoverType toType = toLC;
		LandCoverType fromType = fromLC;

		// reverse direction if negative
		if (change < 0) {
			change = -change;
			toType = fromLC;
			fromType = toLC;
		}

		double shortfall = allocAllLandCrops(newLandUseRaster, keys, toType, fromType, change);

		if (shortfall > 0.00001) {
			LogWriter.printlnError("This should never happen, due to GAMS constraint. Not able to incorporate all changes: from " + fromLC + " to " + toLC + " " + locId + ": " + shortfall);
		}
	}

	private double allocAllLandCrops(RasterSet<LandUseItem> newLandUseRaster, Set<RasterKey> keys, LandCoverType toType, LandCoverType fromType, double change) {
		if (DEBUG) LogWriter.println("allocAllLandCrops: from " + fromType + " to " + toType + " change " + change);
		if (change < 0.00001)
			return 0;

		double totalShortfall = 0;
		Set<RasterKey> keysWithSpace = new HashSet<RasterKey>();
		double totalUnprotectedLC = 0;
		
		for (RasterKey key : keys) {
			LandUseItem newLandUseItem = newLandUseRaster.get(key);
			totalUnprotectedLC += newLandUseItem.getUnprotectedLandCoverArea(fromType);
		}

		for (RasterKey key : keys) {
			//if (DEBUG) LogWriter.println("Processing raster key " + key);
			LandUseItem newLandUseItem = newLandUseRaster.get(key);

			if (newLandUseItem!=null) {
				double cellChange = (totalUnprotectedLC > 0) ? change * newLandUseItem.getUnprotectedLandCoverArea(fromType)/totalUnprotectedLC : change / keys.size();
				double shortfall = newLandUseItem.moveAreas(toType, fromType, cellChange);
				shortfall += newLandUseItem.moveUnprotectedArea(toType, fromType, cellChange);
				if (shortfall == 0)
					keysWithSpace.add(key);
				else
					totalShortfall += shortfall;
				
				//LogWriter.println(key.toString() + "|" + cellChange);
			}
		}

		if (DEBUG) LogWriter.println("allocAllLandCrops: from " + fromType + " to " + toType + " change " + change + " totalShortfall " + totalShortfall);
		return totalShortfall;
	}

	@SuppressWarnings("serial")
	private GamsLocationInput convertFromRaster(GamsRasterInput rasterInputData) {
		// as a first attempt only going to look at pasture and cereal yields, assuming other yields will be correlated to one or the other.
		YieldRaster yieldRaster = rasterInputData.getYields();
		RasterSet<LandUseItem> landUseRaster = rasterInputData.getPreviousLandUses();

		RasterSet<IrrigationItem> irrigRaster = rasterInputData.getIrrigationData();
		
		RasterSet<CarbonFluxItem> carbonFluxRaster = rasterInputData.getCarbonFluxes();
		
		RasterSet<WoodYieldItem> woodYieldRaster = rasterInputData.getWoodYields();

		// look for inconsistencies
		for (Entry<RasterKey, YieldResponsesItem> entry : yieldRaster.entrySet()) {

			YieldResponsesItem yresp = entry.getValue();
			RasterKey key = entry.getKey();
			CropType crop = CropType.WHEAT;

			if (yresp != null && (yresp.getYield(YieldType.FERT_MAX_IRRIG_MAX, crop) < yresp.getYield(YieldType.FERT_MAX_NO_IRRIG, crop) || yresp.getYield(YieldType.FERT_MAX_IRRIG_MAX, crop) < yresp.getYield(YieldType.NO_FERT_IRRIG_MAX, crop))) {
				logWarningWithCoord("Inconsistency F only:" + yresp.getYield(YieldType.FERT_MAX_NO_IRRIG, crop) + ", I only" + yresp.getYield(YieldType.NO_FERT_IRRIG_MAX, crop) + ", max " + yresp.getYield(YieldType.FERT_MAX_IRRIG_MAX, crop) + " at ", key, yieldRaster);
			}
		}

		// Aggregation
		LazyTreeMap<Integer, YieldResponsesItem> aggregatedYields = new LazyTreeMap<Integer, YieldResponsesItem>() { 
			protected YieldResponsesItem createValue() { return new YieldResponsesItem(); }
		};
		LazyTreeMap<Integer, LandUseItem> aggregatedAreas = new LazyTreeMap<Integer, LandUseItem>() { 
			protected LandUseItem createValue() { return new LandUseItem(); }
		};
		LazyTreeMap<Integer, IrrigationItem> aggregatedIrrigCosts = new LazyTreeMap<Integer, IrrigationItem>() { 
			protected IrrigationItem createValue() { return new IrrigationItem(); }
		};
		LazyTreeMap<Integer, CarbonFluxItem> aggregatedCarbonFluxes = new LazyTreeMap<Integer, CarbonFluxItem>() { 
			protected CarbonFluxItem createValue() { return new CarbonFluxItem(); }
		};
		LazyTreeMap<Integer, WoodYieldItem> aggregatedWoodYields = new LazyTreeMap<Integer, WoodYieldItem>() { 
			protected WoodYieldItem createValue() { return new WoodYieldItem(); }
		};

		int countFound = 0, countMissing = 0;

		for (Entry<RasterKey, YieldResponsesItem> entry : yieldRaster.entrySet()) {
			YieldResponsesItem yresp = entry.getValue();
			RasterKey key = entry.getKey();

			if (yresp == null) {
				countMissing++;
				//		logErrorWithCoord("No YieldResponsesItem for ", key, yieldRaster);
			}
			else {
				countFound++;
				//logErrorWithCoord("YieldResponsesItem found for ", key, yieldRaster);

				LandUseItem landUseItem  = landUseRaster.get(key);
				if (landUseItem == null) {
					LogWriter.printlnError("GamsRasterOptimiser: Can't find cropAreas for " + key);
					continue;
				}

				IrrigationItem irrigItem = irrigRaster.get(key);
				
				CarbonFluxItem carbonFluxItem = carbonFluxRaster.get(key);
				WoodYieldItem woodYieldItem = woodYieldRaster.get(key);

				int clusterId = mapping.get(key).getInt();

				YieldResponsesItem aggYResp = aggregatedYields.lazyGet(clusterId);
				LandUseItem aggLandUse = aggregatedAreas.lazyGet(clusterId);
				IrrigationItem aggIrig = aggregatedIrrigCosts.lazyGet(clusterId);
				
				CarbonFluxItem aggCFlux = aggregatedCarbonFluxes.lazyGet(clusterId);
				WoodYieldItem aggWYield = aggregatedWoodYields.lazyGet(clusterId);
				
				/*
				if (clusterId == 4 && rasterInputData.getTimestep().getYear() == 2011) {
					int foo = 1;
				}
				*/
				
				// TODO disabled protected area update
				//landUseItem.updateSuitableArea(rasterInputData.getTimestep().getYear());
				double suitableAreaThisTime  = landUseItem.getSuitableArea();
				double suitableAreaSoFar = aggLandUse.getSuitableArea();
				
				// Irrigation cost
				if (irrigItem!= null) {
					aggIrig.setIrrigCost( aggregateMean(aggIrig.getIrrigCost(), suitableAreaSoFar, irrigItem.getIrrigCost(), suitableAreaThisTime));
					aggIrig.setIrrigConstraint(aggregateMean(aggIrig.getIrrigConstraint(), suitableAreaSoFar, irrigItem.getIrrigConstraint(), suitableAreaThisTime));
					//LogWriter.println(id + ", " + key + ", " + avgIrigCost.getIrrigCost() + ", from " + areaSoFar + ", " + suitableAreaThisTime + ", " + irrigCost.getIrrigCost());
				}
				
				/*
				if (carbonFluxItem == null) {
					LogWriter.println("" + rasterInputData.getCarbonFluxes().getXCoordin(key));
					LogWriter.println("" + rasterInputData.getCarbonFluxes().getYCoordin(key));
				}
				*/
				
				/*
				if (clusterId == 16) {
					int foo =1;
				} else {
					continue;
				}
				*/
				
				// Aggregate carbon fluxes and wood yields
				for (LandCoverType prevLC : LandCoverType.getConvertibleTypes()) {
					for (LandCoverType newLC : LandCoverType.getConvertibleTypes()) {
						// Carbon flux
						double carbonFlux;
						if (carbonFluxItem != null) {
							carbonFlux = carbonFluxItem.getCarbonFlux(prevLC, newLC);
						} else {
							carbonFlux = 0.0; // if missing data assume 0 TODO is this right?
						}

						if (!aggCFlux.checkForKeys(prevLC, newLC)) { // if not seen yet
							aggCFlux.setCarbonFlux(prevLC, newLC, carbonFlux);
						} else {
							aggCFlux.setCarbonFlux(prevLC, newLC, aggregateMean(aggCFlux.getCarbonFlux(prevLC, newLC), suitableAreaSoFar, 
									carbonFlux, suitableAreaThisTime));
						}
						// Wood yield
						double woodYield;
						if (woodYieldItem != null) {
							woodYield = woodYieldItem.getWoodYield(prevLC, newLC);
						} else {
							woodYield = 0.0; // if missing data assume 0
						}

						if (!aggWYield.checkForKeys(prevLC, newLC)) { // if not seen yet
							aggWYield.setWoodYield(prevLC, newLC, woodYield);
						} else {
							aggWYield.setWoodYield(prevLC, newLC, aggregateMean(aggWYield.getWoodYield(prevLC, newLC), suitableAreaSoFar, 
									woodYield, suitableAreaThisTime));
						}
					}
				}

				
				// Crops yields and area fractions
				for (CropType crop : CropType.getNonMeatTypes()) {
					if (!crop.equals(CropType.SETASIDE)) {
						if (irrigItem!= null) {							
							double irrigMax = irrigItem.getMaxIrrigAmount(crop);

							if (Double.isNaN(irrigMax))
								logWarningWithCoord("Can't find irrig max amount for ", key, yieldRaster, crop);
							else 
								aggIrig.setMaxIrrigAmount(crop, aggregateMean(aggIrig.getMaxIrrigAmount(crop), suitableAreaSoFar, irrigMax, suitableAreaThisTime));
						}

						for (YieldType yieldType : YieldType.values()) {
							double y = yresp.getYield(yieldType, crop);
							if (Double.isNaN(y))
								logWarningWithCoord("Problem getting yield for type:" + yieldType + ", ", key, yieldRaster, crop);
							else {
								double yieldSoFar = aggYResp.getYield(yieldType, crop);
								double updatedYield = aggregateMean(yieldSoFar, suitableAreaSoFar, y, suitableAreaThisTime);
								aggYResp.setYield(yieldType, crop, updatedYield);
							}
						}

						double s = yresp.getShockRate(crop);
						if (!Double.isNaN(s)) {
							double shockSoFar = aggYResp.getShockRate(crop);
							double updatedShock = aggregateMean(shockSoFar, suitableAreaSoFar, s, suitableAreaThisTime);
							aggYResp.setShockRate(crop, updatedShock);
						}
					}

					double areaSoFar = aggLandUse.getCropArea(crop);
					double areaThisTime = landUseItem.getCropArea(crop);
					double updateCropland = (landUseItem.getLandCoverArea(LandCoverType.CROPLAND) + aggLandUse.getLandCoverArea(LandCoverType.CROPLAND));

					if (updateCropland > 0)
						aggLandUse.setCropFraction(crop, (areaSoFar + areaThisTime) / updateCropland);

					Intensity intensityThisTime = landUseItem.getIntensity(crop);
					aggLandUse.setIntensity(crop, intensityThisTime);  // intensity currently is always the same within a location, so can just that any.  If this changes need to average here.
				}

				// Unavailable Fraction
				double unavailAreaThisTime = landUseItem.getUnavailableArea();
				double unavailAreaSoFar = aggLandUse.getUnavailableArea();
				aggLandUse.setUnavailableArea(unavailAreaSoFar + unavailAreaThisTime);

				// Protected areas
				double protectedThisTime = landUseItem.getProtectedArea();
				double protectedSoFar = aggLandUse.getProtectedArea();
				aggLandUse.setProtectedArea(protectedSoFar + protectedThisTime);

				//Suitable areas
				aggLandUse.setSuitableArea(suitableAreaSoFar + suitableAreaThisTime);

				
				// Land covers areas
				for (LandCoverType landType : LandCoverType.values()) {
					double unprotectedAreaThisTime = landUseItem.getUnprotectedLandCoverArea(landType);
					double unprotectedAreaSoFar = aggLandUse.getUnprotectedLandCoverArea(landType);
					aggLandUse.setUnprotectedLandCoverArea(landType, unprotectedAreaSoFar + unprotectedAreaThisTime);
					
					double totalAreaThisTime = landUseItem.getLandCoverArea(landType);
					double totalAreaSoFar = aggLandUse.getLandCoverArea(landType);
					aggLandUse.setLandCoverArea(landType, totalAreaSoFar + totalAreaThisTime);
				}			
				
			}
		}	

		LogWriter.println("YieldResponsesItem: " + rasterInputData.getCountryInput().getCountry() + ", countFound=" + countFound + ", countMissing=" + countMissing);

		checkedTotalAreas(landUseRaster, "before");
		checkedTotalAreas(aggregatedAreas, "after");
		
		return new GamsLocationInput(rasterInputData.getTimestep(), aggregatedYields, aggregatedAreas, aggregatedIrrigCosts, 
				aggregatedCarbonFluxes, aggregatedWoodYields, rasterInputData.getMinimumLandCover(), rasterInputData.getTimberYield(), rasterInputData.getNaturalWoodYieldFactor(),
				rasterInputData.getConversionCosts(), rasterInputData.getCountryInput());
	}

	private void logWarningWithCoord(String message, RasterKey key, YieldRaster yieldRaster, CropType crop) {
		logWarningWithCoord(message + "crop:" + crop + ", ", key, yieldRaster);
	}

	private void logWarningWithCoord(String message, RasterKey key, YieldRaster yieldRaster) {
		LogWriter.printlnWarning(message + key + ", x:" + yieldRaster.getXCoordin(key) + ", y:" + yieldRaster.getYCoordin(key));
	}

	private double aggregateMean(double aggV, double aggArea, double newV, double newArea) {
		if (Double.isNaN(aggV))
			return newV;

		if (newArea == 0)
			return aggV;

		return (aggV*aggArea + newV*newArea) / (aggArea + newArea);
	}
}
