package ac.ed.lurg.demand;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.country.CompositeCountry;
import ac.ed.lurg.country.CompositeCountryManager;
import ac.ed.lurg.country.SingleCountry;
import ac.ed.lurg.types.CommodityType;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.utils.LogWriter;
import ac.ed.lurg.utils.StringTabularReader;

public class CerealFractionsManager {
	protected CompositeCountryManager compositeCountryManager;
	private StringTabularReader cerealFractionReader;

	public CerealFractionsManager(CompositeCountryManager compositeCountryManager) {
		this.compositeCountryManager = compositeCountryManager;
		cerealFractionReader = new StringTabularReader(",", new String[]{"Country", "plumDemandItem", "Item", "food"});
		cerealFractionReader.read(ModelConfig.BASE_DEMAND_FRACT_FILE);
	}

	public Map<CommodityType, Map<CropType, Double>> getBaseDemandFracts(CompositeCountry cc) {
		Map<CommodityType, Map<CropType, Double>> fracts = new HashMap<CommodityType, Map<CropType, Double>>();
		
		for (CommodityType commodity : CommodityType.getAllFoodItems()) {
			Collection<CropType> crops = commodity.getCropTypes();
			Map<CropType, Double> fractsCrops;
			
			if (crops.size() == 1) {
				fractsCrops = new HashMap<CropType, Double>();
				fractsCrops.put(crops.iterator().next(), 1.0);
			}
			else {
				fractsCrops = getBaseDemandFracts(cc, commodity);
			}
			
			fracts.put(commodity, fractsCrops);
		}
		
		return fracts;
	}
	
	private Map<CropType, Double> getBaseDemandFracts(CompositeCountry cc, CommodityType commodity) {

		Map<CropType, Double> commodityDemandMap = new HashMap<CropType, Double>();
		for (SingleCountry c : compositeCountryManager.getAllForCompositeCountry(cc)) {
			Map<CropType, Double> singleCMap = getCommodityFraction(c, commodity);

			for (CropType crop : commodity.getCropTypes()) {
				Double newDemandD = singleCMap.get(crop);
				Double prevDemandD = commodityDemandMap.get(crop);
				double newDemand = getDoubleOrZero(newDemandD) + getDoubleOrZero(prevDemandD);
				
				commodityDemandMap.put(crop, newDemand);
			}
		}
		
		double totalCommodityDemand = 0;
		for (Double d : commodityDemandMap.values())
			totalCommodityDemand += getDoubleOrZero(d);

		Map<CropType, Double> commodityFractionMap = new HashMap<CropType, Double>();
		for (Entry<CropType, Double> entry : commodityDemandMap.entrySet()) {
			commodityFractionMap.put(entry.getKey(), getDoubleOrZero(entry.getValue())/totalCommodityDemand);
		}

		return commodityFractionMap;
	}

	private double getDoubleOrZero(Double d) {
		return (d==null ? 0.0 : d.doubleValue());
	}
	
	private Map<CropType, Double> getCommodityFraction(SingleCountry c, CommodityType commodity) {
		Map<CropType, Double> foodDemandMap = new HashMap<CropType, Double>();

		for (CropType crop : commodity.getCropTypes()) {
			Map<String, String> queryMap = new HashMap<String, String>();
			queryMap.put("Country", c.getCountryName());
			queryMap.put("Item", crop.getFaoName());
			queryMap.put("plumDemandItem", commodity.getFaoName());

			try {
				Map<String, String> row = cerealFractionReader.querySingle(queryMap);
				String commodityS = row.get("food");
				Double commodityAmt = Double.valueOf(commodityS);
				foodDemandMap.put(crop, commodityAmt);
			} catch (Exception e) {
				LogWriter.println("Problem finding " + commodity.getGamsName() + " demand: " + crop.getFaoName() + ", "
						+ c.getCountryName());
			}
		}
		return foodDemandMap;
	}
}
